{-- Haskell98 --}

-- Number-parameterized vectors: fixed-precision decimal numbers, 
-- type-constructor-based encoding (the size is the phantom type)

module FixedDecT (
		  D0(..),D1(..),D2(..),D3(..),D4(..),
		  D5(..),D6(..),D7(..),D8(..),D9(..),
		  Digit(..), NonZeroDigit, Card(..),
		  Vec,		       -- Vec data constructor is not exported!
		  listVec,
		  vec,
		  vlength_t, vlength,
		  vupdate,
		  velems, vat, vzipWith
		  )
where

import Data.Array

-- Types and constructors for individual digits
-- Depending on the context, D5 stands either for the type for numeral 5,
-- or for the data constructor for numeral 5

data D0 = D0
data D1 = D1
data D2 = D2
data D3 = D3
data D4 = D4
data D5 = D5
data D6 = D6
data D7 = D7
data D8 = D8
data D9 = D9

class Digit d where	-- class of digits
    d2num:: (Num a) => d -> a	-- convert to a number

instance Digit D0 where d2num _ = 0
instance Digit D1 where d2num _ = 1
instance Digit D2 where d2num _ = 2
instance Digit D3 where d2num _ = 3
instance Digit D4 where d2num _ = 4
instance Digit D5 where d2num _ = 5
instance Digit D6 where d2num _ = 6
instance Digit D7 where d2num _ = 7
instance Digit D8 where d2num _ = 8
instance Digit D9 where d2num _ = 9

class Digit d => NonZeroDigit d
instance NonZeroDigit D1
instance NonZeroDigit D2
instance NonZeroDigit D3
instance NonZeroDigit D4
instance NonZeroDigit D5
instance NonZeroDigit D6
instance NonZeroDigit D7
instance NonZeroDigit D8
instance NonZeroDigit D9

-- Class of non-negative numbers
class Card c where
    c2num:: (Num a) => c -> a -- convert to a number

-- Single-digit numbers are non-negative numbers
instance Card D0 where c2num _ = 0
instance Card D1 where c2num _ = 1
instance Card D2 where c2num _ = 2
instance Card D3 where c2num _ = 3
instance Card D4 where c2num _ = 4
instance Card D5 where c2num _ = 5
instance Card D6 where c2num _ = 6
instance Card D7 where c2num _ = 7
instance Card D8 where c2num _ = 8
instance Card D9 where c2num _ = 9

t12::(a,b)   -> a; t12 = undefined
t22::(a,b)   -> b; t22 = undefined
t13::(a,b,c) -> a; t13 = undefined
t23::(a,b,c) -> b; t23 = undefined
t33::(a,b,c) -> c; t33 = undefined

-- two-digit numbers are also non-negative numbers
instance (NonZeroDigit d1,Digit d2) => Card (d1,d2) where
    c2num c = 10*(d2num $ t12 c) + (d2num $ t22 c)

-- three-digit numbers are also non-negative numbers
instance (NonZeroDigit d1,Digit d2,Digit d3) => Card (d1,d2,d3) where
    c2num c = 100*(d2num $ t13 c) + 10*(d2num $ t23 c) + (d2num $ t33 c)

-- The data constructor Vec is NOT to be exported
newtype Vec size a = Vec (Array Int a) deriving Show

vec:: Card size => size -> a -> Vec size a
vec size elem = Vec $ listArray (0,(c2num size)-1) $ repeat elem

-- The following is an internal function. It does not check
-- that the list `elems' has the `size' elements.
-- The function is used internally when we have proven that |elems| is 
-- indeed size or when it's OK to truncate elems to size
listVec':: Card size => size -> [a] -> Vec size a
listVec' size elems = Vec $ listArray (0,(c2num size)-1) elems

-- The following is an exported function. 
listVec:: Card size => size -> [a] -> Vec size a
listVec size elems | not (c2num size == length elems) =
		       error "listVec: static/dynamic sizes mismatch"
listVec size elems = listVec' size elems

-- gives static, compile-time size
vlength_t:: Vec size a -> size
vlength_t _ = undefined

-- gives run-time size
vlength:: Vec size a -> Int
vlength (Vec arr) = let (0,last) = bounds arr in last+1

vat::  Vec size a -> Int -> a
vat (Vec arr) i = arr ! i

velems:: Vec size a -> [a]
velems (Vec v) = elems v

vupdate:: Vec size a -> [(Int,a)] -> Vec size a
vupdate (Vec arr) ul = Vec $ arr // ul

-- The type of vzipWith assures the safety side-condition of listVec'
vzipWith:: Card size => 
   (a->b->c) -> Vec size a -> Vec size b -> Vec size c
vzipWith f va vb = 
   listVec' (vlength_t va) $ zipWith f (velems va) (velems vb)

-- Tests

sample_v = listVec D2 [True,False]

v12c = listVec (D1,D2) $ take 12 ['a'..'z']
v12i = listVec (D1,D2) [1..12]
v13i = listVec (D1,D3) [1..13]

-- :t v12c

test1 = vlength v12c; test1' = vlength_t v12c
test2 = [vat v12c 0, vat v12c 1, vat v12c 2]

test3 = vzipWith (\c i -> (toEnum $ fromEnum c + fromIntegral i)::Char) 
	         v12c v12i
test4 = vzipWith (+) v12i v12i

-- The following leads to a type error, as expected:
-- test5 = vzipWith (+) v12i v13i

{--
Vec (array (0,1) [(0,True),(1,False)])
*FixedDecT> :type sample_v
Vec D2 Bool
*FixedDecT> vec (D1,D2) 'a'
Vec (array (0,11) [(0,'a'),(1,'a'),(2,'a'),(3,'a'),(4,'a'),(5,'a'),(6,'a'),(7,'a'),(8,'a'),(9,'a'),(10,'a'),(11,'a')])
*FixedDecT> :t vec (D1,D2) 'a'
Vec (D1, D2) Char

--}
